import pytest

from pandas.tests.extension import base

from cyberpandas.mac_array import MACArray, MACType


@pytest.fixture
def dtype():
    return MACType()


@pytest.fixture
def data():
    return MACArray(list(range(100)))


@pytest.fixture
def data_missing():
    return MACArray([0, 1])


@pytest.fixture(params=['data', 'data_missing'])
def all_data(request, data, data_missing):
    """Parametrized fixture giving 'data' and 'data_missing'"""
    if request.param == 'data':
        return data
    elif request.param == 'data_missing':
        return data_missing


@pytest.fixture
def data_for_sorting():
    return MACArray([10, 2 ** 64 - 1, 1])


@pytest.fixture
def data_missing_for_sorting():
    return MACArray([2 ** 64 - 1, 0, 1])


@pytest.fixture
def data_for_grouping():
    b = 1
    a = 2 ** 32 + 1
    c = 2 ** 32 + 10
    return MACArray([
        b, b, 0, 0, a, a, b, c
    ])


@pytest.fixture
def na_cmp():
    """Binary operator for comparing NA values.

    Should return a function of two arguments that returns
    True if both arguments are (scalar) NA for your type.

    By defult, uses ``operator.or``
    """
    return lambda x, y: int(x) == int(y) == 0


@pytest.fixture
def na_value():
    return MACType.na_value


class TestDtype(base.BaseDtypeTests):
    pass


class TestInterface(base.BaseInterfaceTests):
    pass


class TestConstructors(base.BaseConstructorsTests):
    pass


class TestReshaping(base.BaseReshapingTests):
    @pytest.mark.skip(reason='Pandas inferrs us as int64.')
    def test_concat_mixed_dtypes(self):
        pass


class TestGetitem(base.BaseGetitemTests):
    pass


class TestMissing(base.BaseMissingTests):
    pass


class TestMethods(base.BaseMethodsTests):
    @pytest.mark.xfail(reason='upstream')
    def test_value_counts(data, dropna):
        pass



"""Custom IP address dtype / block for pandas"""

from .ip_array import (
    IPType,
    IPArray,
    IPAccessor,
)
from .ip_methods import ip_range
from .parser import to_ipaddress
from .mac_array import MACType, MACArray

from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass

del get_distribution
del DistributionNotFound


__all__ = [
    '__version__',
    'IPAccessor',
    'IPArray',
    'IPType',
    'MACArray',
    'MACType',
    'ip_range',
    'to_ipaddress',
]



import abc
import collections
import ipaddress

import six
import numpy as np
import pandas as pd
from pandas.api.extensions import ExtensionDtype

from ._accessor import (DelegatedMethod, DelegatedProperty,
                        delegated_method)
from ._utils import combine, pack, unpack
from .base import NumPyBackedExtensionArrayMixin
from .common import _U8_MAX, _IPv4_MAX
from .parser import _to_ipaddress_pyint, _as_ip_object

# -----------------------------------------------------------------------------
# Extension Type
# -----------------------------------------------------------------------------


@six.add_metaclass(abc.ABCMeta)
class IPv4v6Base(object):
    """Metaclass providing a common base class for the two scalar IP types."""
    pass


IPv4v6Base.register(ipaddress.IPv4Address)
IPv4v6Base.register(ipaddress.IPv6Address)


class IPType(ExtensionDtype):
    name = 'ip'
    type = IPv4v6Base
    kind = 'O'
    _record_type = np.dtype([('hi', '>u8'), ('lo', '>u8')])
    na_value = ipaddress.IPv4Address(0)

    @classmethod
    def construct_from_string(cls, string):
        if string == cls.name:
            return cls()
        else:
            raise TypeError("Cannot construct a '{}' from "
                            "'{}'".format(cls, string))

# -----------------------------------------------------------------------------
# Extension Container
# -----------------------------------------------------------------------------


class IPArray(NumPyBackedExtensionArrayMixin):
    """Holder for IP Addresses.

    IPArray is a container for IPv4 or IPv6 addresses. It satisfies pandas'
    extension array interface, and so can be stored inside
    :class:`pandas.Series` and :class:`pandas.DataFrame`.

    See :ref:`usage` for more.
    """
    # A note on the internal data layout. IPv6 addresses require 128 bits,
    # which is more than a uint64 can store. So we use a NumPy structured array
    # with two fields, 'hi', 'lo' to store the data. Each field is a uint64.
    # The 'hi' field contains upper 64 bits. The think this is correct since
    # all IP traffic is big-endian.
    __array_priority__ = 1000
    _dtype = IPType()
    _itemsize = 16
    ndim = 1
    can_hold_na = True

    def __init__(self, values):
        from .parser import _to_ip_array

        values = _to_ip_array(values)  # TODO: avoid potential copy
        self.data = values

    @classmethod
    def from_pyints(cls, values):
        # type: (T.Sequence[int]) -> 'IPArray'
        """Construct an IPArray from a sequence of Python integers.

        This can be useful for representing IPv6 addresses, which may
        be larger than 2**64.

        Parameters
        ----------
        values : Sequence
            Sequence of Python integers.

        Examples
        --------
        >>> IPArray.from_pyints([0, 10, 2 ** 64 + 1])
        IPArray(['0.0.0.1', '0.0.0.2', '0.0.0.3', '0:0:0:1::'])
        """
        return cls(_to_ipaddress_pyint(values))

    @classmethod
    def from_bytes(cls, bytestring):
        """Create an IPArray from a bytestring.

        Parameters
        ----------
        bytestring : bytes
            Note that bytestring is a Python 3-style string of bytes,
            not a sequences of bytes where each element represents an
            IPAddress.

        Returns
        -------
        IPArray

        Examples
        --------
        >>> arr = IPArray([10, 20])
        >>> buf = arr.to_bytes()
        >>> buf
        b'\x00\x00\...x00\x02'
        >>> IPArray.from_bytes(buf)
        IPArray(['0.0.0.10', '0.0.0.20'])

        See Also
        --------
        to_bytes
        from_pyints
        """
        data = np.frombuffer(bytestring, dtype=IPType._record_type)
        return cls._from_ndarray(data)

    @classmethod
    def _from_ndarray(cls, data, copy=False):
        """Zero-copy construction of an IPArray from an ndarray.

        Parameters
        ----------
        data : ndarray
            This should have IPType._record_type dtype
        copy : bool, default False
            Whether to copy the data.

        Returns
        -------
        ExtensionArray
        """
        if copy:
            data = data.copy()
        new = IPArray([])
        new.data = data
        return new

    @property
    def _as_u8(self):
        """A 2-D view on our underlying data, for bit-level manipulation."""
        return self.data.view("<u8").reshape(-1, 1)

    # -------------------------------------------------------------------------
    # Properties
    # -------------------------------------------------------------------------
    @property
    def na_value(self):
        """The missing value sentinal for IP Addresses.

        The address ``'0.0.0.0'`` is used.

        Examples
        --------
        >>> IPArray([]).na_value
        IPv4Address('0.0.0.0')
        """
        return self.dtype.na_value

    def take(self, indices, allow_fill=False, fill_value=None):
        # Can't use pandas' take yet
        # 1. axis
        # 2. I don't know how to do the reshaping correctly.
        indices = np.asarray(indices, dtype='int')

        if allow_fill and fill_value is None:
            fill_value = unpack(pack(int(self.na_value)))
        elif allow_fill and not isinstance(fill_value, tuple):
            fill_value = unpack(pack(int(fill_value)))

        if allow_fill:
            mask = (indices == -1)
            if not len(self):
                if not (indices == -1).all():
                    msg = "Invalid take for empty array. Must be all -1."
                    raise IndexError(msg)
                else:
                    # all NA take from and empty array
                    took = (np.full((len(indices), 2), fill_value, dtype='>u8')
                              .reshape(-1).astype(self.dtype._record_type))
                    return self._from_ndarray(took)
            if (indices < -1).any():
                msg = ("Invalid value in 'indicies'. Must be all >= -1 "
                       "for 'allow_fill=True'")
                raise ValueError(msg)

        took = self.data.take(indices)
        if allow_fill:
            took[mask] = fill_value

        return self._from_ndarray(took)

    # -------------------------------------------------------------------------
    # Interfaces
    # -------------------------------------------------------------------------

    def __repr__(self):
        formatted = self._format_values()
        return "IPArray({!r})".format(formatted)

    def _format_values(self):
        formatted = []
        # TODO: perf
        for i in range(len(self)):
            hi, lo = self.data[i]
            if lo == -1:
                formatted.append("NA")
            elif hi == 0 and lo <= _IPv4_MAX:
                formatted.append(ipaddress.IPv4Address._string_from_ip_int(
                    int(lo)))
            elif hi == 0:
                formatted.append(ipaddress.IPv6Address._string_from_ip_int(
                    int(lo)))
            else:
                # TODO:
                formatted.append(ipaddress.IPv6Address._string_from_ip_int(
                    (int(hi) << 64) + int(lo)))
        return formatted

    @staticmethod
    def _box_scalar(scalar):
        return ipaddress.ip_address(combine(*scalar))

    @property
    def _parser(self):
        from .parser import to_ipaddress
        return to_ipaddress

    def __setitem__(self, key, value):
        from .parser import to_ipaddress

        value = to_ipaddress(value).data
        self.data[key] = value

    def __iter__(self):
        return iter(self.to_pyipaddress())

    # ------------------------------------------------------------------------
    # Serializaiton / Export
    # ------------------------------------------------------------------------

    def to_pyipaddress(self):
        """Convert the array to a list of scalar IP Adress objects.

        Returns
        -------
        addresses : List
            Each element of the list will be an :class:`ipaddress.IPv4Address`
            or :class:`ipaddress.IPv6Address`, depending on the size of that
            element.

        See Also
        --------
        IPArray.to_pyints

        Examples
        ---------
        >>> IPArray(['192.168.1.1', '2001:db8::1000']).to_pyipaddress()
        [IPv4Address('192.168.1.1'), IPv6Address('2001:db8::1000')]
        """
        import ipaddress
        return [ipaddress.ip_address(x) for x in self._format_values()]

    def to_pyints(self):
        """Convert the array to a list of Python integers.

        Returns
        -------
        addresses : List[int]
            These will be Python integers (not NumPy), which are unbounded in
            size.

        See Also
        --------
        IPArray.to_pyipaddresses
        IPArray.from_pyints

        Examples
        --------
        >>> IPArray(['192.168.1.1', '2001:db8::1000']).to_pyints()
        [3232235777, 42540766411282592856903984951653830656]
        """
        return [combine(*map(int, x)) for x in self.data]

    def to_bytes(self):
        """Serialize the IPArray as a Python bytestring.

        This and :meth:IPArray.from_bytes is the fastest way to roundtrip
        serialize and de-serialize an IPArray.

        See Also
        --------
        IPArray.from_bytes

        Examples
        --------
        >>> arr = IPArray([10, 20])
        >>> arr.to_bytes()
        b'\x00\x00\...x00\x02'
        """
        return self.data.tobytes()

    # ------------------------------------------------------------------------
    # Ops
    # ------------------------------------------------------------------------

    def __eq__(self, other):
        # TDOO: scalar ipaddress
        if not isinstance(other, IPArray):
            return NotImplemented
        mask = self.isna() | other.isna()
        result = self.data == other.data
        result[mask] = False
        return result

    def __lt__(self, other):
        # TDOO: scalar ipaddress
        if not isinstance(other, IPArray):
            return NotImplemented
        mask = self.isna() | other.isna()
        result = ((self.data['hi'] <= other.data['hi']) &
                  (self.data['lo'] < other.data['lo']))
        result[mask] = False
        return result

    def __le__(self, other):
        if not isinstance(other, IPArray):
            return NotImplemented
        mask = self.isna() | other.isna()
        result = ((self.data['hi'] <= other.data['hi']) &
                  (self.data['lo'] <= other.data['lo']))
        result[mask] = False
        return result

    def __gt__(self, other):
        if not isinstance(other, IPArray):
            return NotImplemented
        return other < self

    def __ge__(self, other):
        if not isinstance(other, IPArray):
            return NotImplemented
        return other <= self

    def equals(self, other):
        if not isinstance(other, IPArray):
            raise TypeError("Cannot compare 'IPArray' "
                            "to type '{}'".format(type(other)))
        # TODO: missing
        return (self.data == other.data).all()

    def _values_for_factorize(self):
        return self.astype(object), ipaddress.IPv4Address(0)

    def isna(self):
        """Indicator for whether each element is missing.

        The IPAddress 0 is used to indecate missing values.

        Examples
        --------
        >>> IPArray(['0.0.0.0', '192.168.1.1']).isna()
        array([ True, False])
        """
        ips = self.data
        return (ips['lo'] == 0) & (ips['hi'] == 0)

    def isin(self, other):
        """Check whether elements of `self` are in `other`.

        Comparison is done elementwise.

        Parameters
        ----------
        other : str or sequences
            For ``str`` `other`, the argument is attempted to
            be converted to an :class:`ipaddress.IPv4Network` or
            a :class:`ipaddress.IPv6Network` or an :class:`IPArray`.
            If all those conversions fail, a TypeError is raised.

            For a sequence of strings, the same conversion is attempted.
            You should not mix networks with addresses.

            Finally, other may be an ``IPArray`` of addresses to compare to.

        Returns
        -------
        contained : ndarray
            A 1-D boolean ndarray with the same length as self.

        Examples
        --------
        Comparison to a single network

        >>> s = IPArray(['192.168.1.1', '255.255.255.255'])
        >>> s.isin('192.168.1.0/24')
        array([ True, False])

        Comparison to many networks
        >>> s.isin(['192.168.1.0/24', '192.168.2.0/24'])
        array([ True, False])

        Comparison to many IP Addresses

        >>> s.isin(['192.168.1.1', '192.168.1.2', '255.255.255.1']])
        array([ True, False])
        """
        box = (isinstance(other, str) or
               not isinstance(other, (IPArray, collections.Sequence)))
        if box:
            other = [other]

        networks = []
        addresses = []

        if not isinstance(other, IPArray):
            for net in other:
                net = _as_ip_object(net)
                if isinstance(net, (ipaddress.IPv4Network,
                                    ipaddress.IPv6Network)):
                    networks.append(net)
                if isinstance(net, (ipaddress.IPv4Address,
                                    ipaddress.IPv6Address)):
                    addresses.append(ipaddress.IPv6Network(net))
        else:
            addresses = other

        # Flatten all the addresses
        addresses = IPArray(addresses)  # TODO: think about copy=False

        mask = np.zeros(len(self), dtype='bool')
        for network in networks:
            mask |= self._isin_network(network)

        # no... we should flatten this.
        mask |= self._isin_addresses(addresses)
        return mask

    def _isin_network(self, other):
        # type: (Union[ipaddress.IPv4Network,ipaddress.IPv6Network]) -> ndarray
        """Check whether an array of addresses is contained in a network."""
        # A network is bounded below by 'network_address' and
        # above by 'broadcast_address'.
        # IPArray handles comparisons between arrays of addresses, and NumPy
        # handles broadcasting.
        net_lo = type(self)([other.network_address])
        net_hi = type(self)([other.broadcast_address])

        return (net_lo <= self) & (self <= net_hi)

    def _isin_addresses(self, other):
        """Check whether elements of self are present in other."""
        from pandas.core.algorithms import isin
        # TODO(factorize): replace this
        return isin(self, other)

    # ------------------------------------------------------------------------
    # IP Specific
    # ------------------------------------------------------------------------

    @property
    def is_ipv4(self):
        """Indicator for whether each address fits in the IPv4 space."""
        # TODO: NA should be NA
        ips = self.data
        return (ips['hi'] == 0) & (ips['lo'] < _U8_MAX)

    @property
    def is_ipv6(self):
        """Indicator for whether each address requires IPv6."""
        ips = self.data
        return (ips['hi'] > 0) | (ips['lo'] > _U8_MAX)

    @property
    def version(self):
        """IP version (4 or 6)."""
        return np.where(self.is_ipv4, 4, 6)

    @property
    def is_multicast(self):
        """Indiciator for whether each address is multicast."""
        pyips = self.to_pyipaddress()
        return np.array([ip.is_multicast for ip in pyips])

    @property
    def is_private(self):
        """Indiciator for whether each address is private."""
        pyips = self.to_pyipaddress()
        return np.array([ip.is_private for ip in pyips])

    @property
    def is_global(self):
        """Indiciator for whether each address is global."""
        pyips = self.to_pyipaddress()
        return np.array([ip.is_global for ip in pyips])

    @property
    def is_unspecified(self):
        """Indiciator for whether each address is unspecified."""
        pyips = self.to_pyipaddress()
        return np.array([ip.is_unspecified for ip in pyips])

    @property
    def is_reserved(self):
        """Indiciator for whether each address is reserved."""
        pyips = self.to_pyipaddress()
        return np.array([ip.is_reserved for ip in pyips])

    @property
    def is_loopback(self):
        """Indiciator for whether each address is loopback."""
        pyips = self.to_pyipaddress()
        return np.array([ip.is_loopback for ip in pyips])

    @property
    def is_link_local(self):
        """Indiciator for whether each address is link local."""
        pyips = self.to_pyipaddress()
        return np.array([ip.is_link_local for ip in pyips])

    @property
    def packed(self):
        """Bytestring of the IP addresses

        Each address takes 16 bytes. IPv4 addresses are prefixed
        by zeros.
        """
        # TODO: I wonder if that should be post-fixed by 0s.
        return self.data.tobytes()

    def _apply_mask(self, op, v4_prefixlen, v6_prefixlen):
        """Apply a netmask or hostmask"""
        self = self.copy()
        is_v4 = self.is_ipv4
        v4_net = getattr(
            ipaddress.ip_network(u'0.0.0.0/{}'.format(v4_prefixlen)),
            op)
        v4_mask = IPArray([v4_net])
        self.data[is_v4] = v4_mask.data

        v6_net = getattr(
            ipaddress.ip_network(u'0::0/{}'.format(v6_prefixlen)),
            op)
        v6_mask = IPArray([v6_net])
        self.data[~is_v4] = v6_mask.data
        return self

    def netmask(self, v4_prefixlen=32, v6_prefixlen=128):
        """Compute an array of netmasks for an array of IP addresses.

        Note that this is a method, rather than a property, to support
        taking `v4_prefixlen` and `v6_prefixlen` as arguments.

        Parameters
        ----------
        v4_prefixlen : int, default 32
            Length of the network prefix, in bits, for IPv4 addresses
        v6_prefixlen : int, default 128
            Lnegth of the network prefix, in bits, for IPv6 addresses

        Returns
        -------
        IPArray

        See Also
        --------
        IPArray.hostmask

        Examples
        --------
        >>> arr = ip.IPArray(['192.0.0.0', '1:1::'])
        >>> arr.netmask(v4_prefixlen=16, v6_prefixlen=32)
        IPArray(['255.255.0.0', 'ffff:ffff::'])
        """
        return self._apply_mask('netmask', v4_prefixlen, v6_prefixlen)

    def hostmask(self, v4_prefixlen=32, v6_prefixlen=128):
        """Compute an array of hostmasks for an array of IP addresses.

        Parameters
        ----------
        v4_prefixlen : int, default 32
            Length of the network prefix, in bits, for IPv4 addresses
        v6_prefixlen : int, default 128
            Lnegth of the network prefix, in bits, for IPv6 addresses

        Returns
        -------
        IPArray

        See Also
        --------
        IPArray.netmask

        Examples
        --------
        >>> arr = ip.IPArray(['192.0.0.0', '1:1::'])
        >>> arr.netmask(v4_prefixlen=16, v6_prefixlen=32)
        IPArray(['0.0.255.255', '::ffff:ffff:ffff:ffff:ffff:ffff'])
        """
        return self._apply_mask('hostmask', v4_prefixlen, v6_prefixlen)

    def mask(self, mask):
        """Apply a host or subnet mask.

        Parameters
        ----------
        mask : IPArray
            The host or subnet mask to be applied

        Returns
        -------
        masked : IPArray

        See Also
        --------
        netmask
        hostmask

        Examples
        --------
        >>> arr = IPArray(['216.003.128.12', '192.168.100.1'])
        >>> mask = arr.netmask(v4_prefixlen=24)
        >>> mask
        IPArray(['255.255.255.0', '255.255.255.0'])
        >>> arr.mask(mask)
        IPArray(['216.3.128.0', '192.168.100.0'])
        """
        mask = type(self)(mask)
        a = self._as_u8
        b = mask._as_u8
        masked = np.bitwise_and(a, b).ravel().view(self.dtype._record_type)
        return type(self)(masked)


# -----------------------------------------------------------------------------
# Accessor
# -----------------------------------------------------------------------------

@pd.api.extensions.register_series_accessor("ip")
class IPAccessor:

    is_ipv4 = DelegatedProperty("is_ipv4")
    is_ipv6 = DelegatedProperty("is_ipv6")
    version = DelegatedProperty("version")
    is_multicast = DelegatedProperty("is_multicast")
    is_private = DelegatedProperty("is_private")
    is_global = DelegatedProperty("is_global")
    is_unspecified = DelegatedProperty("is_unspecified")
    is_reserved = DelegatedProperty("is_reserved")
    is_loopback = DelegatedProperty("is_loopback")
    is_link_local = DelegatedProperty("is_link_local")

    isna = DelegatedMethod("isna")
    to_pyints = DelegatedMethod("to_pyints")

    def __init__(self, obj):
        self._validate(obj)
        self._data = obj.values
        self._index = obj.index
        self._name = obj.name

    @staticmethod
    def _validate(obj):
        if not is_ipaddress_type(obj):
            raise AttributeError("Cannot use 'ip' accessor on objects of "
                                 "dtype '{}'.".format(obj.dtype))

    def isin(self, other):
        return delegated_method(self._data.isin, self._index,
                                self._name, other)

    def netmask(self, v4_prefixlen=32, v6_prefixlen=128):
        return delegated_method(self._data.netmask, self._index,
                                self._name, v4_prefixlen, v6_prefixlen)

    def hostmask(self, v4_prefixlen=32, v6_prefixlen=128):
        return delegated_method(self._data.hostmask, self._index,
                                self._name, v4_prefixlen, v6_prefixlen)

    def mask(self, other):
        return delegated_method(self._data.mask, self._index, self._name,
                                other)


def is_ipaddress_type(obj):
    t = getattr(obj, 'dtype', obj)
    try:
        return isinstance(t, IPType) or issubclass(t, IPType)
    except Exception:
        return False



import operator

import numpy as np

from pandas.core.arrays import ExtensionArray


class NumPyBackedExtensionArrayMixin(ExtensionArray):
    @property
    def dtype(self):
        """The dtype for this extension array, IPType"""
        return self._dtype

    @classmethod
    def _from_sequence(cls, scalars):
        return cls(scalars)

    @classmethod
    def _from_factorized(cls, values, original):
        return cls(values)

    @property
    def shape(self):
        return (len(self.data),)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, *args):
        result = operator.getitem(self.data, *args)
        if isinstance(result, tuple):
            return self._box_scalar(result)
        elif result.ndim == 0:
            return self._box_scalar(result.item())
        else:
            return type(self)(result)

    def setitem(self, indexer, value):
        """Set the 'value' inplace.
        """
        # I think having a separate than __setitem__ is good
        # since we have to return here, but __setitem__ doesn't.
        self[indexer] = value
        return self

    @property
    def nbytes(self):
        return self._itemsize * len(self)

    def _formatting_values(self):
        return np.array(self._format_values(), dtype='object')

    def copy(self, deep=False):
        return type(self)(self.data.copy())

    @classmethod
    def _concat_same_type(cls, to_concat):
        return cls(np.concatenate([array.data for array in to_concat]))

    def tolist(self):
        return self.data.tolist()

    def argsort(self, axis=-1, kind='quicksort', order=None):
        return self.data.argsort()

    def unique(self):
        # type: () -> ExtensionArray
        # https://github.com/pandas-dev/pandas/pull/19869
        _, indices = np.unique(self.data, return_index=True)
        data = self.data.take(np.sort(indices))
        return self._from_ndarray(data)



import ipaddress

import numpy as np
from pandas.api.types import is_list_like

from ._utils import pack, unpack


def to_ipaddress(values):
    """Convert values to IPArray

    Parameters
    ----------
    values : int, str, bytes, or sequence of those

    Returns
    -------
    addresses : IPArray

    Examples
    --------
    Parse strings
    >>> to_ipaddress(['192.168.1.1',
    ...               '2001:0db8:85a3:0000:0000:8a2e:0370:7334'])
    <IPArray(['192.168.1.1', '0:8a2e:370:7334:2001:db8:85a3:0'])>

    Or integers
    >>> to_ipaddress([3232235777,
                      42540766452641154071740215577757643572])
    <IPArray(['192.168.1.1', '0:8a2e:370:7334:2001:db8:85a3:0'])>

    Or packed binary representations
    >>> to_ipaddress([b'\xc0\xa8\x01\x01',
                      b' \x01\r\xb8\x85\xa3\x00\x00\x00\x00\x8a.\x03ps4'])
    <IPArray(['192.168.1.1', '0:8a2e:370:7334:2001:db8:85a3:0'])>
    """
    from . import IPArray

    if not is_list_like(values):
        values = [values]

    return IPArray(_to_ip_array(values))


def _to_ip_array(values):
    from .ip_array import IPType, IPArray

    if isinstance(values, IPArray):
        return values.data

    if (isinstance(values, np.ndarray) and
            values.ndim == 1 and
            np.issubdtype(values.dtype, np.integer)):
        # We assume we're given the low bits here.
        values = values.astype("u8")
        values = np.asarray(values, dtype=IPType._record_type)
        values['hi'] = 0

    elif not (isinstance(values, np.ndarray) and
              values.dtype == IPType._record_type):
        values = _to_int_pairs(values)
    return np.atleast_1d(np.asarray(values, dtype=IPType._record_type))


def _to_int_pairs(values):
    if isinstance(values, (str, bytes, int)):
        values = ipaddress.ip_address(values)._ip
        return unpack(pack(values))
    elif isinstance(values, np.ndarray) and values.dtype != object:
        if values.ndim != 2:
            raise ValueError("'values' should be a 2-D when passing a "
                             "NumPy array.")
    elif isinstance(values, tuple) and len(values) == 2:
        # like IPArray((0, 0))
        # which isn't IPArray([0, 0])
        pass
    elif all(isinstance(x, tuple) for x in values):
        # TODO: not great
        pass
    else:
        values = [ipaddress.ip_address(v)._ip for v in values]
        values = [unpack(pack(v)) for v in values]
    return values


def _to_ipaddress_pyint(values):
    from .ip_array import IPType

    values2 = [unpack(pack(x)) for x in values]
    return np.atleast_1d(np.asarray(values2, dtype=IPType._record_type))


def _as_ip_object(val):
    """Attempt to parse 'val' as any IP object.

    Attempts to parse as these in order:

    - IP Address (v4 or v6)
    - IP Network (v4 or v6)
    """
    try:
        return ipaddress.ip_address(val)
    except ValueError:
        pass

    try:
        return ipaddress.ip_network(val)
    except ValueError:
        raise ValueError("Could not parse {} is an address or "
                         "network".format(val))



from .common import _IPv4_MAX


def is_ipv4(value):
    # type: (Any) -> bool
    if isinstance(value, str):
        return value.count(".") == 3
    elif isinstance(value, bytes):
        pass
    elif isinstance(value, int):
        return value < _IPv4_MAX
    else:
        return False


def is_ipv6(value):
    # type: (Any) -> bool
    if isinstance(value, str):
        return value.count(":") == 7



import operator

import numpy as np
import pandas as pd
import pandas.util.testing as tm
import pytest

import cyberpandas as ip


@pytest.fixture
def series():
    return pd.Series(ip.IPArray.from_pyints([0, 1, 2]))


@pytest.fixture
def frame():
    return pd.DataFrame({"A": ip.IPArray.from_pyints([0, 1, 2]),
                         "B": [0, 1, 2],
                         "C": ip.IPArray.from_pyints([0, 1, 2])})


@pytest.fixture(params=['series', 'frame'])
def obj(request, series, frame):
    if request.param == 'series':
        return series
    elif request.param == 'frame':
        return frame


# -----
# Tests
# -----
@pytest.mark.parametrize('method', [
    operator.methodcaller('head'),
    operator.methodcaller('rename', str),
])
def test_works_generic(obj, method):
    method(obj)


@pytest.mark.parametrize('method', [
    operator.methodcaller('info'),
])
def test_works_frame(frame, method):
    method(frame)


def test__take(frame):
    return frame._take([0], axis=0)


def test_iloc_series(series):
    series.iloc[slice(None)]
    series.iloc[0]
    series.iloc[[0]]
    series.iloc[[0, 1]]


def test_iloc_frame(frame):
    frame.iloc[:, 0]
    frame.iloc[:, [0]]
    frame.iloc[:, [0, 1]]
    frame.iloc[:, [0, 2]]

    frame.iloc[0, 0]
    frame.iloc[0, [0]]
    frame.iloc[0, [0, 1]]
    frame.iloc[0, [0, 2]]

    frame.iloc[[0], 0]
    frame.iloc[[0], [0]]
    frame.iloc[[0], [0, 1]]
    frame.iloc[[0], [0, 2]]


def test_loc_series(series):
    series.loc[:]
    series.loc[0]
    series.loc[1]
    series.loc[[0, 1]]


def test_loc_frame(frame):
    frame.loc[:, 'A']
    frame.loc[:, ['A']]
    frame.loc[:, ['A', 'B']]
    frame.loc[:, ['A', 'C']]

    frame.loc[0, 'A']
    frame.loc[0, ['A']]
    frame.loc[0, ['A', 'B']]
    frame.loc[0, ['A', 'C']]

    frame.loc[[0], 'A']
    frame.loc[[0], ['A']]
    frame.loc[[0], ['A', 'B']]
    frame.loc[[0], ['A', 'C']]


def test_reindex(frame):
    result = frame.reindex([0, 10])
    expected = pd.DataFrame({"A": ip.IPArray.from_pyints([0, 0]),
                             "B": [0, np.nan],
                             "C": ip.IPArray.from_pyints([0, 0])},
                            index=[0, 10])
    tm.assert_frame_equal(result, expected)


def test_isna(series):
    expected = pd.Series([True, False, False], index=series.index,
                         name=series.name)
    result = pd.isna(series)
    tm.assert_series_equal(result, expected)

    result = series.isna()
    tm.assert_series_equal(result, expected)


def test_isna_frame(frame):
    result = frame.isna()
    expected = pd.DataFrame({"A": [True, False, False],
                             "B": [False, False, False],
                             "C": [True, False, False]})
    tm.assert_frame_equal(result, expected)


@pytest.mark.xfail(reason="Not implemented")
def test_fillna():
    result = pd.Series(ip.IPArray([1, 0])).fillna(method='ffill')
    expected = pd.Series(ip.IPArray([1, 1]))
    tm.assert_series_equal(result, expected)


@pytest.mark.xfail(reason="Not implemented")
def test_dropna():
    missing = pd.Series(ip.IPArray([1, 0]))
    result = missing.dropna()
    expected = pd.Series(ip.IPArray([1]))
    tm.assert_series_equal(result, expected)

    result = missing.to_frame().dropna()
    expected = expected.to_frame()
    tm.assert_frame_equal(result, expected)



from collections import Iterable

import numpy as np
import six

from pandas.api.extensions import ExtensionDtype, take

from .base import NumPyBackedExtensionArrayMixin


class MACType(ExtensionDtype):
    """Dtype for MAC Address Data."""
    name = 'mac'
    # type is long for Py2 and int for py3
    type = six.integer_types[-1]
    kind = 'u'
    na_value = 0  # TODO: Check this.

    @classmethod
    def construct_from_string(cls, string):
        if string == cls.name:
            return cls()
        else:
            raise TypeError("Cannot construct a '{}' from "
                            "'{}'".format(cls, string))


class MACArray(NumPyBackedExtensionArrayMixin):
    """Array for MAC Address data.

    * https://en.wikipedia.org/wiki/MAC_address
    * https://tools.ietf.org/html/rfc5342
    """
    # What type(s) do we support?
    # MAC-48 or EUI-64?
    _dtype = MACType()
    _itemsize = 8
    ndim = 1
    can_hold_na = True

    def __init__(self, values, copy=True):
        # TODO: parse hex / strings
        self.data = np.array(values, dtype='uint64', copy=copy)

    @classmethod
    def _from_ndarray(cls, data, copy=False):
        return cls(data, copy=copy)

    @property
    def na_value(self):
        return self.dtype.na_value

    def __repr__(self):
        formatted = self._format_values()
        return "MACArray({!r})".format(formatted)

    def _format_values(self):
        return [_format(x) for x in self.data]

    @staticmethod
    def _box_scalar(scalar):
        return scalar

    def __setitem__(self, key, value):
        value = to_macaddress(value)
        self.data[key] = value

    def __iter__(self):
        return iter(self.data.tolist())

    def __lt__(self, other):
        return self.data < other

    def __le__(self, other):
        return self.data <= other

    def __eq__(self, other):
        return self.data == other

    def __ge__(self, other):
        return other <= self

    def __gt__(self, other):
        return other < self

    def equals(self, other):
        if not isinstance(other, type(self)):
            raise TypeError
        return (self.data == other.data).all()

    def _values_for_factorize(self):
        # Should hit pandas' UInt64Hashtable
        return self, 0

    def isna(self):
        return (self.data == 0)

    @property
    def _parser(self):
        return lambda x: x

    def take(self, indexer, allow_fill=False, fill_value=None):
        if fill_value is None:
            fill_value = 0
        took = take(self.data, indexer, allow_fill=allow_fill,
                    fill_value=fill_value)
        return type(self)(took)

    def _formatting_values(self):
        return np.array(self._format_values(), dtype='object')

    @classmethod
    def _concat_same_type(cls, to_concat):
        return cls(np.concatenate([array.data for array in to_concat]))

    def take_nd(self, indexer, allow_fill=True, fill_value=None):
        return self.take(indexer, allow_fill=allow_fill, fill_value=fill_value)

    def copy(self, deep=False):
        return type(self)(self.data.copy())


def _format(mac):
    # https://stackoverflow.com/a/36883363/1889400
    mac_hex = "{:012x}".format(mac)
    mac_str = ":".join(mac_hex[i:i+2] for i in range(0, len(mac_hex), 2))
    return mac_str


def _parse(mac):
    # https://stackoverflow.com/a/36883363/1889400
    mac_int = int(mac.replace(":", "").replace("-", ""), 16)
    return mac_int


def to_macaddress(addresses):
    if (isinstance(addresses, six.string_types) or
            not isinstance(addresses, Iterable)):
        addresses = [addresses]

    addresses = [_parse(mac) if isinstance(mac, six.string_types) else mac
                 for mac in addresses]
    return np.array(addresses, dtype='u8')



import pandas as pd


def delegated_method(method, index, name, *args, **kwargs):
    return pd.Series(method(*args, **kwargs), index, name=name)


class Delegated:
    # Descriptor for delegating attribute access to from
    # a Series to an underlying array

    def __init__(self, name):
        self.name = name

    def __get__(self, obj, type=None):
        index = object.__getattribute__(obj, '_index')
        name = object.__getattribute__(obj, '_name')
        result = self._get_result(obj)
        return pd.Series(result, index, name=name)


class DelegatedProperty(Delegated):
    def _get_result(self, obj, type=None):
        return getattr(object.__getattribute__(obj, '_data'), self.name)


class DelegatedMethod(Delegated):
    def __get__(self, obj, type=None):
        index = object.__getattribute__(obj, '_index')
        name = object.__getattribute__(obj, '_name')
        method = getattr(object.__getattribute__(obj, '_data'), self.name)
        return delegated_method(method, index, name)
